# Copyright (c) 2013-2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import copy

from oslo_config import cfg
from oslo_log import log as logging

from neutron.extensions import portbindings
from neutron.plugins.ml2 import driver_api as api

from networking_odl.common import config as odl_conf
from networking_odl.common import journal
from networking_odl.db import db
from networking_odl.ml2 import network_topology

LOG = logging.getLogger(__name__)


class OpenDaylightMechanismDriver(api.MechanismDriver):
    """OpenDaylight Python Driver for Neutron.

    This code is the backend implementation for the OpenDaylight ML2
    MechanismDriver for OpenStack Neutron.
    """

    def initialize(self):
        LOG.debug("Initializing OpenDaylight ML2 driver")
        cfg.CONF.register_opts(odl_conf.odl_opts, "ml2_odl")
        self.vif_details = {portbindings.CAP_PORT_FILTER: True}
        self.journal = journal.OpendaylightJournalThread()
        self._network_topology = network_topology.NetworkTopologyManager()

    @journal.call_thread_on_end
    def create_network_precommit(self, context):
        db.create_pending_row(context._plugin_context.session, 'network',
                              context.current['id'], 'create', context.current)

    @journal.call_thread_on_end
    def create_subnet_precommit(self, context):
        db.create_pending_row(context._plugin_context.session, 'subnet',
                              context.current['id'], 'create', context.current)

    @journal.call_thread_on_end
    def create_port_precommit(self, context):
        dbcontext = context._plugin_context
        groups = [context._plugin.get_security_group(dbcontext, sg)
                  for sg in context.current['security_groups']]
        new_context = copy.deepcopy(context.current)
        new_context['security_groups'] = groups
        # NOTE(yamahata): work around for port creation for router
        # tenant_id=''(empty string) is passed when port is created
        # by l3 plugin internally for router.
        # On the other hand, ODL doesn't accept empty string for tenant_id.
        # In that case, deduce tenant_id from network_id for now.
        # Right fix: modify Neutron so that don't allow empty string
        # for tenant_id even for port for internal use.
        # TODO(yamahata): eliminate this work around when neutron side
        # is fixed
        # assert port['tenant_id'] != ''
        if ('tenant_id' not in context.current or
                context.current['tenant_id'] == ''):
            tenant_id = context._network_context._network['tenant_id']
            new_context['tenant_id'] = tenant_id
        db.create_pending_row(context._plugin_context.session, 'port',
                              context.current['id'], 'create', new_context)

    @journal.call_thread_on_end
    def update_network_precommit(self, context):
        db.create_pending_row(context._plugin_context.session, 'network',
                              context.current['id'], 'update', context.current)

    @journal.call_thread_on_end
    def update_subnet_precommit(self, context):
        db.create_pending_row(context._plugin_context.session, 'subnet',
                              context.current['id'], 'update', context.current)

    @journal.call_thread_on_end
    def update_port_precommit(self, context):
        port = context._plugin.get_port(context._plugin_context,
                                        context.current['id'])
        dbcontext = context._plugin_context
        new_context = copy.deepcopy(context.current)
        groups = [context._plugin.get_security_group(dbcontext, sg)
                  for sg in port['security_groups']]
        new_context['security_groups'] = groups
        # Add the network_id in for validation
        new_context['network_id'] = port['network_id']
        # NOTE(yamahata): work around for port creation for router
        # tenant_id=''(empty string) is passed when port is created
        # by l3 plugin internally for router.
        # On the other hand, ODL doesn't accept empty string for tenant_id.
        # In that case, deduce tenant_id from network_id for now.
        # Right fix: modify Neutron so that don't allow empty string
        # for tenant_id even for port for internal use.
        # TODO(yamahata): eliminate this work around when neutron side
        # is fixed
        # assert port['tenant_id'] != ''
        if ('tenant_id' not in context.current or
                context.current['tenant_id'] == ''):
            port['tenant_id'] = context._network_context._network['tenant_id']
        db.create_pending_row(context._plugin_context.session, 'port',
                              context.current['id'], 'update', new_context)

    @journal.call_thread_on_end
    def delete_network_precommit(self, context):
        db.create_pending_row(context._plugin_context.session, 'network',
                              context.current['id'], 'delete', context.current)

    @journal.call_thread_on_end
    def delete_subnet_precommit(self, context):
        db.create_pending_row(context._plugin_context.session, 'subnet',
                              context.current['id'], 'delete', context.current)

    @journal.call_thread_on_end
    def delete_port_precommit(self, context):
        db.create_pending_row(context._plugin_context.session, 'port',
                              context.current['id'], 'delete', context.current)

    def bind_port(self, port_context):
        """Set binding for a valid segments

        """
        return self._network_topology.bind_port(port_context)



# Copyright (c) 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from sqlalchemy import asc
from sqlalchemy import func
from sqlalchemy import or_

from networking_odl.common import constants as odl_const
from networking_odl.db import models

from neutron.db import api as db_api

from oslo_db import api as oslo_db_api


def _check_for_pending_or_processing_ops(session, object_uuid):
    q = session.query(models.OpendaylightJournal).filter(
        or_(models.OpendaylightJournal.state == 'pending',
            models.OpendaylightJournal.state == 'processing'),
        models.OpendaylightJournal.object_uuid == object_uuid)
    return session.query(q.exists()).scalar()


def get_all_db_rows(session):
    return session.query(models.OpendaylightJournal).all()


def get_all_db_rows_by_state(session, state):
    return session.query(models.OpendaylightJournal).filter_by(
        state=state).all()


def get_untried_db_row_with_lock(session):
    return session.query(models.OpendaylightJournal).filter_by(
        state='pending', retry_count=0).with_for_update().first()


def get_oldest_pending_db_row_with_lock(session):
    row = session.query(models.OpendaylightJournal).filter_by(
        state='pending').order_by(
        asc(models.OpendaylightJournal.last_retried)).with_for_update().first()
    if row:
        update_pending_db_row_processing(session, row)

    return row


@oslo_db_api.wrap_db_retry(max_retries=db_api.MAX_RETRIES,
                           retry_on_request=True)
def update_pending_db_row_processing(session, row):
    row.state = 'processing'
    session.merge(row)
    session.flush()


@oslo_db_api.wrap_db_retry(max_retries=db_api.MAX_RETRIES,
                           retry_on_request=True)
def update_pending_db_row_retry(session, row, retry_count):
    if row.retry_count >= retry_count:
        row.state = 'failed'
    else:
        row.retry_count = row.retry_count + 1
        row.state = 'pending'
    session.merge(row)
    session.flush()


@oslo_db_api.wrap_db_retry(max_retries=db_api.MAX_RETRIES,
                           retry_on_request=True)
def update_processing_db_row_passed(session, row):
    row.state = 'completed'
    session.merge(row)
    session.flush()


@oslo_db_api.wrap_db_retry(max_retries=db_api.MAX_RETRIES,
                           retry_on_request=True)
def update_db_row_pending(session, row):
    row.state = 'pending'
    session.merge(row)
    session.flush()


# This function is currently not used.
# Deleted resources are marked as 'deleted' in the database.
@oslo_db_api.wrap_db_retry(max_retries=db_api.MAX_RETRIES,
                           retry_on_request=True)
def delete_row(session, row=None, row_id=None):
    if row_id:
        row = session.query(models.OpendaylightJournal).filter_by(
            id=row_id).one()
    if row:
        session.delete(row)
        session.flush()


@oslo_db_api.wrap_db_retry(max_retries=db_api.MAX_RETRIES,
                           retry_on_request=True)
def create_pending_row(session, object_type, object_uuid,
                       operation, data):
    row = models.OpendaylightJournal(object_type=object_type,
                                     object_uuid=object_uuid,
                                     operation=operation, data=data,
                                     created_at=func.now(), state='pending')
    session.add(row)
    # Session flush required for L3 events. NOOP for L2 events since ML2
    # precommit calls are made inside database session transaction.
    session.flush()


def validate_network_operation(session, object_uuid, operation, data):
    """Validate the network operation based on dependencies.

    Validate network operation depending on whether it's dependencies
    are still in 'pending' or 'processing' state. e.g.
    """
    return True


def validate_subnet_operation(session, object_uuid, operation, data):
    """Validate the subnet operation based on dependencies.

    Validate subnet operation depending on whether it's dependencies
    are still in 'pending' or 'processing' state. e.g.
    """
    if operation in ('create', 'update'):
        network_id = data['network_id']
        # Check for pending or processing network operations
        if _check_for_pending_or_processing_ops(session, network_id):
            return False
    elif operation == 'delete':
        # TODO(asomya):Check for pending port operations
        pass

    return True


def validate_port_operation(session, object_uuid, operation, data):
    """Validate port operation based on dependencies.

    Validate port operation depending on whether it's dependencies
    are still in 'pending' or 'processing' state. e.g.
    """
    if operation in ('create', 'update'):
        network_id = data['network_id']
        # Check for pending or processing network operations
        ops = _check_for_pending_or_processing_ops(session, network_id)
        # Check for pending subnet operations.
        for fixed_ip in data['fixed_ips']:
            ip_ops = _check_for_pending_or_processing_ops(
                session, fixed_ip['subnet_id'])
            ops = ops or ip_ops

        if ops:
            return False

    return True

VALIDATION_MAP = {
    odl_const.ODL_NETWORK: validate_network_operation,
    odl_const.ODL_SUBNET: validate_subnet_operation,
    odl_const.ODL_PORT: validate_port_operation,
}



# Copyright (c) 2015-2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import mock

from neutron.tests import base


class DietTestCase(base.DietTestCase):

    def patch(self, target, name, *args, **kwargs):
        context = mock.patch.object(target, name, *args, **kwargs)
        patch = context.start()
        self.addCleanup(context.stop)
        return patch



# Copyright (c) 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import abc
import six

from networking_odl.common import constants as odl_const
from networking_odl.common import utils as odl_utils


@six.add_metaclass(abc.ABCMeta)
class ResourceFilterBase(object):
    @staticmethod
    @abc.abstractmethod
    def filter_create_attributes(resource):
        pass

    @staticmethod
    @abc.abstractmethod
    def filter_update_attributes(resource):
        pass


class NetworkFilter(ResourceFilterBase):
    @staticmethod
    def filter_create_attributes(network):
        """Filter out network attributes not required for a create."""
        odl_utils.try_del(network, ['status', 'subnets'])

    @staticmethod
    def filter_update_attributes(network):
        """Filter out network attributes for an update operation."""
        odl_utils.try_del(network, ['id', 'status', 'subnets', 'tenant_id'])


class SubnetFilter(ResourceFilterBase):
    @staticmethod
    def filter_create_attributes(subnet):
        """Filter out subnet attributes not required for a create."""
        pass

    @staticmethod
    def filter_update_attributes(subnet):
        """Filter out subnet attributes for an update operation."""
        odl_utils.try_del(subnet, ['id', 'network_id', 'ip_version', 'cidr',
                          'allocation_pools', 'tenant_id'])


class PortFilter(ResourceFilterBase):
    @classmethod
    def filter_create_attributes(cls, port):
        """Filter out port attributes not required for a create."""
        # TODO(kmestery): Converting to uppercase due to ODL bug
        # https://bugs.opendaylight.org/show_bug.cgi?id=477
        port['mac_address'] = port['mac_address'].upper()
        odl_utils.try_del(port, ['status'])

    @classmethod
    def filter_update_attributes(cls, port):
        """Filter out port attributes for an update operation."""
        odl_utils.try_del(port, ['network_id', 'id', 'status', 'mac_address',
                          'tenant_id', 'fixed_ips'])


class SecurityGroupFilter(ResourceFilterBase):
    @staticmethod
    def filter_create_attributes(sg, context):
        """Filter out security-group attributes not required for a create."""
        pass

    @staticmethod
    def filter_update_attributes(sg, context):
        """Filter out security-group attributes for an update operation."""
        pass


class SecurityGroupRuleFilter(ResourceFilterBase):
    @staticmethod
    def filter_create_attributes(sg_rule, context):
        """Filter out sg-rule attributes not required for a create."""
        pass

    @staticmethod
    def filter_update_attributes(sg_rule, context):
        """Filter out sg-rule attributes for an update operation."""
        pass

FILTER_MAP = {
    odl_const.ODL_NETWORK: NetworkFilter,
    odl_const.ODL_SUBNET: SubnetFilter,
    odl_const.ODL_PORT: PortFilter,
}



# Copyright (c) 2014 Red Hat Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log
from oslo_serialization import jsonutils
from oslo_utils import excutils
import requests


LOG = log.getLogger(__name__)


class OpenDaylightRestClient(object):

    @classmethod
    def create_client(cls):
        return cls(
            cfg.CONF.ml2_odl.url,
            cfg.CONF.ml2_odl.username,
            cfg.CONF.ml2_odl.password,
            cfg.CONF.ml2_odl.timeout)

    def __init__(self, url, username, password, timeout):
        self.url = url
        self.timeout = timeout
        self.auth = (username, password)

    def get(self, urlpath='', data=None):
        return self.request('get', urlpath, data)

    def put(self, urlpath='', data=None):
        return self.request('put', urlpath, data)

    def delete(self, urlpath='', data=None):
        return self.request('delete', urlpath, data)

    def request(self, method, urlpath='', data=None):
        headers = {'Content-Type': 'application/json'}
        url = '/'.join([self.url, urlpath])
        LOG.debug(
            "Sending METHOD (%(method)s) URL (%(url)s) JSON (%(data)s)",
            {'method': method, 'url': url, 'data': data})
        return requests.request(
            method, url=url, headers=headers, data=data, auth=self.auth,
            timeout=self.timeout)

    def sendjson(self, method, urlpath, obj):
        """Send json to the OpenDaylight controller."""
        data = jsonutils.dumps(obj, indent=2) if obj else None
        return self._check_rensponse(self.request(method, urlpath, data))

    def try_delete(self, urlpath):
        rensponse = self.delete(urlpath)
        if rensponse.status_code == requests.codes.not_found:
            # The resource is already removed. ignore 404 gracefully
            LOG.debug("%(urlpath)s doesn't exist", {'urlpath': urlpath})
            return False
        else:
            self._check_rensponse(rensponse)
            return True

    def _check_rensponse(self, rensponse):
        try:
            rensponse.raise_for_status()
        except requests.HTTPError as error:
            with excutils.save_and_reraise_exception():
                LOG.debug("Exception from ODL: %(e)s %(text)s",
                          {'e': error, 'text': rensponse.text}, exc_info=1)
        else:
            LOG.debug("Got response:\n"
                      "(%(response)s)", {'response': rensponse.text})
            return rensponse



# Copyright (c) 2014 Red Hat Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import socket

from oslo_log import log

from networking_odl.common import cache

LOG = log.getLogger(__name__)


def try_del(d, keys):
    """Ignore key errors when deleting from a dictionary."""
    for key in keys:
        try:
            del d[key]
        except KeyError:
            pass


def _fetch_all_addresses_by_hostnames(hostnames):
    for name in hostnames:
        # it uses an ordered dict to avoid duplicates and keep order
        entries = collections.OrderedDict(
            (info[4][0], None) for info in socket.getaddrinfo(name, None))
        for entry in entries:
            yield name, entry


_addresses_by_name_cache = cache.Cache(_fetch_all_addresses_by_hostnames)


def get_addresses_by_name(name, time_to_live=60.0):
    """Gets and caches addresses for given name.

    This is a cached wrapper for function 'socket.getaddrinfo'.

    :returns: a sequence of unique addresses binded to given hostname.
    """

    try:
        results = _addresses_by_name_cache.fetch_all(
            [name], timeout=time_to_live)
        return tuple(address for name, address in results)
    except cache.CacheFetchError as error:
        error.reraise_cause()



# Copyright (c) 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from neutron.tests import base

from networking_odl.common import cache


class TestCache(base.DietTestCase):

    def test_init_with_callable(self):

        def given_fetch_method():
            pass

        cache.Cache(given_fetch_method)

    def test_init_without_callable(self):
        self.assertRaises(TypeError, lambda: cache.Cache(object()))

    def test_fecth_once(self):
        value = 'value'

        given_fetch_method = mock.Mock(return_value=iter([('key', value)]))
        given_cache = cache.Cache(given_fetch_method)

        # When value with key is fetched
        result = given_cache.fetch('key', 60.0)

        # Result is returned
        self.assertIs(value, result)

        # Then fetch method is called once
        given_fetch_method.assert_called_once_with(('key',))

    def test_fecth_with_no_result(self):
        given_fetch_method = mock.Mock(return_value=iter([]))
        given_cache = cache.Cache(given_fetch_method)

        # When value with key is fetched
        try:
            given_cache.fetch('key', 60.0)
        except cache.CacheFetchError as error:
            given_fetch_method.assert_called_once_with(('key',))
            self.assertRaises(KeyError, error.reraise_cause)
        else:
            self.fail('Expecting CacheFetchError to be raised.')

    @mock.patch.object(cache, 'LOG')
    def test_fecth_with_failure(self, logger):
        # pylint: disable=unused-argument

        given_error = RuntimeError("It doesn't work like this!")

        def failing_function(keys):
            raise given_error

        given_fetch_method = mock.Mock(side_effect=failing_function)
        given_cache = cache.Cache(given_fetch_method)

        # When value with key is fetched
        try:
            given_cache.fetch('key', 60.0)
        except cache.CacheFetchError as error:
            given_fetch_method.assert_called_once_with(('key',))
            self.assertRaises(RuntimeError, error.reraise_cause)
        else:
            self.fail('Expecting CacheFetchError to be raised.')
        logger.warning.assert_called_once_with(
            'Error fetching values for keys: %r', "'key'",
            exc_info=(type(given_error), given_error, mock.ANY))

    def test_fecth_again_after_clear(self):
        value1 = 'value1'
        value2 = 'value2'
        given_fetch_method = mock.Mock(
            side_effect=[iter([('key', value1)]),
                         iter([('key', value2)])])
        given_cache = cache.Cache(given_fetch_method)

        # When value with key is fetched
        result1 = given_cache.fetch('key', 60.0)

        # When cache is cleared
        given_cache.clear()

        # When value with same key is fetched again
        result2 = given_cache.fetch('key', 0.0)

        # Then first result is returned
        self.assertIs(value1, result1)

        # Then fetch method is called twice
        self.assertEqual(
            [mock.call(('key',)), mock.call(('key',))],
            given_fetch_method.mock_calls)

        # Then second result is returned
        self.assertIs(value2, result2)

    def test_fecth_again_before_timeout(self):
        value1 = 'value1'
        value2 = 'value2'
        given_fetch_method = mock.Mock(
            side_effect=[iter([('key', value1)]),
                         iter([('key', value2)])])
        given_cache = cache.Cache(given_fetch_method)

        # When value with key is fetched
        result1 = given_cache.fetch('key', 1.0)

        # When value with same key is fetched again and cached entry is not
        # expired
        result2 = given_cache.fetch('key', 0.0)

        # First result is returned
        self.assertIs(value1, result1)

        # Then fetch method is called once
        given_fetch_method.assert_called_once_with(('key',))

        # Then first result is returned twice
        self.assertIs(value1, result2)

    def test_fecth_again_after_timeout(self):
        value1 = 'value1'
        value2 = 'value2'
        given_fetch_method = mock.Mock(
            side_effect=[iter([('key', value1)]),
                         iter([('key', value2)])])
        given_cache = cache.Cache(given_fetch_method)

        # When value with key is fetched
        result1 = given_cache.fetch('key', 0.0)

        # When value with same key is fetched again and cached entry is
        # expired
        result2 = given_cache.fetch('key', 0.0)

        # Then first result is returned
        self.assertIs(value1, result1)

        # Then fetch method is called twice
        self.assertEqual(
            [mock.call(('key',)), mock.call(('key',))],
            given_fetch_method.mock_calls)

        # Then second result is returned
        self.assertIs(value2, result2)

    def test_fecth_two_values_yielding_both_before_timeout(self):
        value1 = 'value1'
        value2 = 'value2'
        given_fetch_method = mock.Mock(
            return_value=iter([('key1', value1),
                               ('key2', value2)]))
        given_cache = cache.Cache(given_fetch_method)

        # When value with key is fetched
        result1 = given_cache.fetch('key1', 60.0)

        # When value with another key is fetched and cached entry is not
        # expired
        result2 = given_cache.fetch('key2', 60.0)

        # Then first result is returned
        self.assertIs(value1, result1)

        # Then fetch method is called once
        given_fetch_method.assert_called_once_with(('key1',))

        # Then second result is returned
        self.assertIs(value2, result2)

    def test_fecth_two_values_yielding_both_after_timeout(self):
        value1 = 'value1'
        value2 = 'value2'
        given_fetch_method = mock.Mock(
            return_value=[('key1', value1), ('key2', value2)])
        given_cache = cache.Cache(given_fetch_method)

        # When value with key is fetched
        result1 = given_cache.fetch('key1', 0.0)

        # When value with another key is fetched and cached entry is
        # expired
        result2 = given_cache.fetch('key2', 0.0)

        # Then first result is returned
        self.assertIs(value1, result1)

        # Then fetch method is called twice
        self.assertEqual(
            [mock.call(('key1',)), mock.call(('key2',))],
            given_fetch_method.mock_calls)

        # Then second result is returned
        self.assertIs(value2, result2)

    def test_fecth_all_with_multiple_entries(self):
        given_fetch_method = mock.Mock(
            return_value=iter([('key', 'value1'),
                               ('key', 'value2')]))
        given_cache = cache.Cache(given_fetch_method)

        # When value with key is fetched
        results = list(given_cache.fetch_all(['key'], 0.0))

        # Then fetch method is once
        given_fetch_method.assert_called_once_with(('key',))

        # Then both results are yield in the right order
        self.assertEqual([('key', 'value1'), ('key', 'value2')], results)

    def test_fecth_all_with_repeated_entries(self):
        entry = ('key', 'value')
        given_fetch_method = mock.Mock(
            return_value=iter([entry, entry, entry]))
        given_cache = cache.Cache(given_fetch_method)

        # When value with key is fetched
        results = list(given_cache.fetch_all(['key'], 0.0))

        # Then fetch method is once
        given_fetch_method.assert_called_once_with(('key',))

        # Then results are yield in the right order
        self.assertEqual([entry, entry, entry], results)



# Copyright (c) 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from neutron.tests import base

from networking_odl.common import cache
from networking_odl.common import utils


class TestGetAddressesByName(base.DietTestCase):

    # pylint: disable=protected-access, unused-argument

    def setUp(self):
        super(TestGetAddressesByName, self).setUp()
        self.clear_cache()
        self.addCleanup(self.clear_cache)
        time = self.patch(
            utils.cache, 'time', clock=mock.Mock(return_value=0.0))
        self.clock = time.clock
        socket = self.patch(utils, 'socket')
        self.getaddrinfo = socket.getaddrinfo

    def patch(self, target, name, *args, **kwargs):
        context = mock.patch.object(target, name, *args, **kwargs)
        mocked = context.start()
        self.addCleanup(context.stop)
        return mocked

    def clear_cache(self):
        utils._addresses_by_name_cache.clear()

    def test_get_addresses_by_valid_name(self):
        self.getaddrinfo.return_value = [
            (2, 1, 6, '', ('127.0.0.1', 0)),
            (2, 2, 17, '', ('127.0.0.1', 0)),
            (2, 3, 0, '', ('127.0.0.1', 0)),
            (2, 1, 6, '', ('10.237.214.247', 0)),
            (2, 2, 17, '', ('10.237.214.247', 0)),
            (2, 3, 0, '', ('10.237.214.247', 0))]

        # When valid host name is requested
        result = utils.get_addresses_by_name('some_host_name')

        # Then correct addresses are returned
        self.assertEqual(('127.0.0.1', '10.237.214.247'), result)

        # Then fetched addresses are cached
        self.assertEqual(result, utils.get_addresses_by_name('some_host_name'))

        # Then addresses are fetched only once
        self.getaddrinfo.assert_called_once_with('some_host_name', None)

    def test_get_addresses_by_valid_name_when_cache_expires(self):
        self.getaddrinfo.return_value = [
            (2, 1, 6, '', ('127.0.0.1', 0)),
            (2, 2, 17, '', ('127.0.0.1', 0)),
            (2, 3, 0, '', ('127.0.0.1', 0)),
            (2, 1, 6, '', ('10.237.214.247', 0)),
            (2, 2, 17, '', ('10.237.214.247', 0)),
            (2, 3, 0, '', ('10.237.214.247', 0))]

        # When valid host name is requested
        result1 = utils.get_addresses_by_name('some_host_name')

        # and after a long time
        self.clock.return_value = 1.0e6

        # When valid host name is requested
        result2 = utils.get_addresses_by_name('some_host_name')

        # Then correct addresses are returned
        self.assertEqual(('127.0.0.1', '10.237.214.247'), result1)
        self.assertEqual(('127.0.0.1', '10.237.214.247'), result2)

        # Then addresses are fetched twice
        self.getaddrinfo.assert_has_calls(
            [mock.call('some_host_name', None),
             mock.call('some_host_name', None)])

    @mock.patch.object(cache, 'LOG')
    def test_get_addresses_by_invalid_name(self, cache_logger):

        # Given addresses resolution is failing
        given_error = RuntimeError("I don't know him!")

        def failing_getaddrinfo(name, service):
            raise given_error

        self.getaddrinfo.side_effect = failing_getaddrinfo

        # When invalid name is requested
        self.assertRaises(
            RuntimeError, utils.get_addresses_by_name, 'some_host_name')

        # When invalid name is requested again
        self.assertRaises(
            RuntimeError, utils.get_addresses_by_name, 'some_host_name')

        # Then result is fetched only once
        self.getaddrinfo.assert_has_calls(
            [mock.call('some_host_name', None)])
        cache_logger.warning.assert_has_calls(
            [mock.call(
                'Error fetching values for keys: %r', "'some_host_name'",
                exc_info=(RuntimeError, given_error, mock.ANY)),
             mock.call(
                'Error fetching values for keys: %r', "'some_host_name'",
                exc_info=(RuntimeError, given_error, mock.ANY))])

    @mock.patch.object(cache, 'LOG')
    def test_get_addresses_failing_when_expired_in_cache(self, cache_logger):
        self.getaddrinfo.return_value = [
            (2, 1, 6, '', ('127.0.0.1', 0)),
            (2, 2, 17, '', ('127.0.0.1', 0)),
            (2, 3, 0, '', ('127.0.0.1', 0)),
            (2, 1, 6, '', ('10.237.214.247', 0)),
            (2, 2, 17, '', ('10.237.214.247', 0)),
            (2, 3, 0, '', ('10.237.214.247', 0))]

        # Given valid result is in chache but expired
        utils.get_addresses_by_name('some_host_name')
        self.clock.return_value = 1.0e6

        # Given addresses resolution is now failing
        given_error = RuntimeError("This is top secret.")

        def failing_getaddrinfo(name, service):
            raise given_error

        self.getaddrinfo.side_effect = failing_getaddrinfo

        self.assertRaises(
            RuntimeError, utils.get_addresses_by_name, 'some_host_name')

        # Then result is fetched more times
        self.getaddrinfo.assert_has_calls(
            [mock.call('some_host_name', None),
             mock.call('some_host_name', None)])
        cache_logger.warning.assert_called_once_with(
            'Error fetching values for keys: %r', "'some_host_name'",
            exc_info=(RuntimeError, given_error, mock.ANY))



# Copyright (c) 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_odl.common import client
from networking_odl.common import constants as odl_const
from networking_odl.common import filters
from networking_odl.common import journal
from networking_odl.db import db
from networking_odl.ml2 import mech_driver_v2

import mock
from oslo_config import cfg
from oslo_serialization import jsonutils
import requests

from neutron.db import api as neutron_db_api
from neutron.plugins.ml2 import config as config
from neutron.plugins.ml2 import plugin
from neutron.tests.unit.plugins.ml2 import test_plugin
from neutron.tests.unit import testlib_api

cfg.CONF.import_group('ml2_odl', 'networking_odl.common.config')

HOST = 'fake-host'
PLUGIN_NAME = 'neutron.plugins.ml2.plugin.Ml2Plugin'
SECURITY_GROUP = '2f9244b4-9bee-4e81-bc4a-3f3c2045b3d7'


class OpenDaylightConfigBase(test_plugin.Ml2PluginV2TestCase):
    def setUp(self):
        super(OpenDaylightConfigBase, self).setUp()
        config.cfg.CONF.set_override('mechanism_drivers',
                                     ['logger', 'opendaylight'], 'ml2')
        config.cfg.CONF.set_override('url', 'http://127.0.0.1:9999', 'ml2_odl')
        config.cfg.CONF.set_override('username', 'someuser', 'ml2_odl')
        config.cfg.CONF.set_override('password', 'somepass', 'ml2_odl')


class OpenDaylightTestCase(OpenDaylightConfigBase):
    def setUp(self):
        super(OpenDaylightTestCase, self).setUp()
        self.port_create_status = 'DOWN'
        self.mech = mech_driver_v2.OpenDaylightMechanismDriver()
        mock.patch.object(journal.OpendaylightJournalThread,
                          'start_odl_sync_thread').start()
        self.mock_sendjson = mock.patch.object(client.OpenDaylightRestClient,
                                               'sendjson').start()
        self.mock_sendjson.side_effect = self.check_sendjson

    def check_sendjson(self, method, urlpath, obj):
        self.assertFalse(urlpath.startswith("http://"))


class OpenDayLightMechanismConfigTests(testlib_api.SqlTestCase):
    def _set_config(self, url='http://127.0.0.1:9999', username='someuser',
                    password='somepass'):
        config.cfg.CONF.set_override('mechanism_drivers',
                                     ['logger', 'opendaylight'],
                                     'ml2')
        config.cfg.CONF.set_override('url', url, 'ml2_odl')
        config.cfg.CONF.set_override('username', username, 'ml2_odl')
        config.cfg.CONF.set_override('password', password, 'ml2_odl')

    def _test_missing_config(self, **kwargs):
        self._set_config(**kwargs)
        self.assertRaises(config.cfg.RequiredOptError,
                          plugin.Ml2Plugin)

    def test_valid_config(self):
        self._set_config()
        plugin.Ml2Plugin()

    def test_missing_url_raises_exception(self):
        self._test_missing_config(url=None)

    def test_missing_username_raises_exception(self):
        self._test_missing_config(username=None)

    def test_missing_password_raises_exception(self):
        self._test_missing_config(password=None)


class OpenDaylightMechanismTestBasicGet(test_plugin.TestMl2BasicGet,
                                        OpenDaylightTestCase):
    pass


class OpenDaylightMechanismTestNetworksV2(test_plugin.TestMl2NetworksV2,
                                          OpenDaylightTestCase):
    pass


class OpenDaylightMechanismTestSubnetsV2(test_plugin.TestMl2SubnetsV2,
                                         OpenDaylightTestCase):
    pass


class OpenDaylightMechanismTestPortsV2(test_plugin.TestMl2PortsV2,
                                       OpenDaylightTestCase):
    pass


class DataMatcher(object):

    def __init__(self, operation, object_type, context):
        self._data = context.current.copy()
        self._object_type = object_type
        filter_cls = filters.FILTER_MAP[object_type]
        attr_filter = getattr(filter_cls, 'filter_%s_attributes' % operation)
        attr_filter(self._data)

    def __eq__(self, s):
        data = jsonutils.loads(s)
        return self._data == data[self._object_type]


class OpenDaylightMechanismDriverTestCase(OpenDaylightConfigBase):
    def setUp(self):
        super(OpenDaylightMechanismDriverTestCase, self).setUp()
        self.db_session = neutron_db_api.get_session()
        self.mech = mech_driver_v2.OpenDaylightMechanismDriver()
        mock.patch.object(journal.OpendaylightJournalThread,
                          'start_odl_sync_thread').start()
        self.mech.initialize()
        self.thread = journal.OpendaylightJournalThread()
        self.addCleanup(self._db_cleanup)

    @staticmethod
    def _get_mock_network_operation_context():
        current = {'status': 'ACTIVE',
                   'subnets': [],
                   'name': 'net1',
                   'provider:physical_network': None,
                   'admin_state_up': True,
                   'tenant_id': 'test-tenant',
                   'provider:network_type': 'local',
                   'router:external': False,
                   'shared': False,
                   'id': 'd897e21a-dfd6-4331-a5dd-7524fa421c3e',
                   'provider:segmentation_id': None}
        context = mock.Mock(current=current)
        context._plugin_context.session = neutron_db_api.get_session()
        return context

    @staticmethod
    def _get_mock_subnet_operation_context():
        current = {'ipv6_ra_mode': None,
                   'allocation_pools': [{'start': '10.0.0.2',
                                         'end': '10.0.1.254'}],
                   'host_routes': [],
                   'ipv6_address_mode': None,
                   'cidr': '10.0.0.0/23',
                   'id': '72c56c48-e9b8-4dcf-b3a7-0813bb3bd839',
                   'name': '',
                   'enable_dhcp': True,
                   'network_id': 'd897e21a-dfd6-4331-a5dd-7524fa421c3e',
                   'tenant_id': 'test-tenant',
                   'dns_nameservers': [],
                   'gateway_ip': '10.0.0.1',
                   'ip_version': 4,
                   'shared': False}
        context = mock.Mock(current=current)
        context._plugin_context.session = neutron_db_api.get_session()
        return context

    @staticmethod
    def _get_mock_port_operation_context():
        current = {'status': 'DOWN',
                   'binding:host_id': '',
                   'allowed_address_pairs': [],
                   'device_owner': 'fake_owner',
                   'binding:profile': {},
                   'fixed_ips': [],
                   'id': '72c56c48-e9b8-4dcf-b3a7-0813bb3bd839',
                   'security_groups': [SECURITY_GROUP],
                   'device_id': 'fake_device',
                   'name': '',
                   'admin_state_up': True,
                   'network_id': 'd897e21a-dfd6-4331-a5dd-7524fa421c3e',
                   'tenant_id': 'test-tenant',
                   'binding:vif_details': {},
                   'binding:vnic_type': 'normal',
                   'binding:vif_type': 'unbound',
                   'mac_address': '12:34:56:78:21:b6'}
        context = mock.Mock(current=current)
        context._plugin.get_security_group = mock.Mock(
            return_value=SECURITY_GROUP)
        context._plugin.get_port = mock.Mock(return_value=current)
        context._plugin_context.session = neutron_db_api.get_session()
        context._network_context = mock.Mock(
            _network=OpenDaylightMechanismDriverTestCase.
            _get_mock_network_operation_context().current)
        return context

    @classmethod
    def _get_mock_operation_context(cls, object_type):
        getter = getattr(cls, '_get_mock_%s_operation_context' % object_type)
        return getter()

    _status_code_msgs = {
        200: '',
        201: '',
        204: '',
        400: '400 Client Error: Bad Request',
        401: '401 Client Error: Unauthorized',
        403: '403 Client Error: Forbidden',
        404: '404 Client Error: Not Found',
        409: '409 Client Error: Conflict',
        501: '501 Server Error: Not Implemented',
        503: '503 Server Error: Service Unavailable',
    }

    def _db_cleanup(self):
        rows = db.get_all_db_rows(self.db_session)
        for row in rows:
            db.delete_row(self.db_session, row=row)

    @classmethod
    def _get_mock_request_response(cls, status_code):
        response = mock.Mock(status_code=status_code)
        response.raise_for_status = mock.Mock() if status_code < 400 else (
            mock.Mock(side_effect=requests.exceptions.HTTPError(
                cls._status_code_msgs[status_code])))
        return response

    def _test_operation(self, method, context, status_code, expected_calls,
                        *args, **kwargs):
        request_response = self._get_mock_request_response(status_code)
        with mock.patch('requests.request',
                        return_value=request_response) as mock_method:
            method(exit_after_run=True)

        if expected_calls:
            mock_method.assert_called_with(
                headers={'Content-Type': 'application/json'},
                auth=(config.cfg.CONF.ml2_odl.username,
                      config.cfg.CONF.ml2_odl.password),
                timeout=config.cfg.CONF.ml2_odl.timeout, *args, **kwargs)
        self.assertEqual(expected_calls, mock_method.call_count)

    def _call_operation_object(self, operation, object_type):
        context = self._get_mock_operation_context(object_type)
        method = getattr(self.mech, '%s_%s_precommit' % (operation,
                                                         object_type))
        method(context)

    def _test_operation_object(self, operation, object_type):
        self._call_operation_object(operation, object_type)

        context = self._get_mock_operation_context(object_type)
        row = db.get_oldest_pending_db_row_with_lock(self.db_session)
        self.assertEqual(operation, row['operation'])
        self.assertEqual(object_type, row['object_type'])
        self.assertEqual(context.current['id'], row['object_uuid'])

    def _test_thread_processing(self, operation, object_type,
                                expected_calls=1):
        http_requests = {odl_const.ODL_CREATE: 'post',
                         odl_const.ODL_UPDATE: 'put',
                         odl_const.ODL_DELETE: 'delete'}
        status_codes = {odl_const.ODL_CREATE: requests.codes.created,
                        odl_const.ODL_UPDATE: requests.codes.ok,
                        odl_const.ODL_DELETE: requests.codes.no_content}

        http_request = http_requests[operation]
        status_code = status_codes[operation]

        self._call_operation_object(operation, object_type)

        context = self._get_mock_operation_context(object_type)
        if operation in [odl_const.ODL_UPDATE, odl_const.ODL_DELETE]:
            url = '%s/%ss/%s' % (config.cfg.CONF.ml2_odl.url,
                                 object_type, context.current['id'])
        else:
            url = '%s/%ss' % (config.cfg.CONF.ml2_odl.url, object_type)

        if operation in [odl_const.ODL_CREATE, odl_const.ODL_UPDATE]:
            kwargs = {
                'url': url,
                'data': DataMatcher(operation, object_type, context)}
        else:
            kwargs = {'url': url, 'data': None}
        with mock.patch.object(self.thread.event, 'wait',
                               return_value=False):
            self._test_operation(self.thread.sync_pending_row, context,
                                 status_code, expected_calls, http_request,
                                 **kwargs)

    def _test_object_type(self, object_type):
        # Add and process create request.
        self._test_thread_processing(odl_const.ODL_CREATE, object_type)
        rows = db.get_all_db_rows_by_state(self.db_session, 'completed')
        self.assertEqual(1, len(rows))

        # Add and process update request. Adds to database.
        self._test_thread_processing(odl_const.ODL_UPDATE, object_type)
        rows = db.get_all_db_rows_by_state(self.db_session, 'completed')
        self.assertEqual(2, len(rows))

        # Add and process update request. Adds to database.
        self._test_thread_processing(odl_const.ODL_DELETE, object_type)
        rows = db.get_all_db_rows_by_state(self.db_session, 'completed')
        self.assertEqual(3, len(rows))

    def _test_object_type_pending_network(self, object_type):
        # Create a network (creates db row in pending state).
        self._call_operation_object(odl_const.ODL_CREATE,
                                    odl_const.ODL_NETWORK)

        # Create object_type database row and process. This results in both
        # the object_type and network rows being processed.
        self._test_thread_processing(odl_const.ODL_CREATE, object_type,
                                     expected_calls=2)

        # Verify both rows are now marked as completed.
        rows = db.get_all_db_rows_by_state(self.db_session, 'completed')
        self.assertEqual(2, len(rows))

    def _test_object_type_processing_network(self, object_type):
        # Create a network (creates db row in pending state).
        self._call_operation_object(odl_const.ODL_CREATE,
                                    odl_const.ODL_NETWORK)

        # Get pending network row and mark as processing so that
        # this row will not be processed by journal thread.
        row = db.get_all_db_rows_by_state(self.db_session, 'pending')
        db.update_pending_db_row_processing(self.db_session, row[0])

        # Create the object_type database row and process.
        # Verify that object create request is not processed because the
        # dependent network row has not been marked as 'completed'.
        self._test_thread_processing(odl_const.ODL_CREATE,
                                     object_type,
                                     expected_calls=0)

        # Verify that network and object rows are still in the database.
        rows = db.get_all_db_rows_by_state(self.db_session, 'processing')
        self.assertEqual(2, len(rows))

    def test_driver(self):
        for operation in [odl_const.ODL_CREATE, odl_const.ODL_UPDATE,
                          odl_const.ODL_DELETE]:
            for object_type in [odl_const.ODL_NETWORK, odl_const.ODL_SUBNET,
                                odl_const.ODL_PORT]:
                self._test_operation_object(operation, object_type)

    def test_network(self):
        self._test_object_type(odl_const.ODL_NETWORK)

    def test_subnet(self):
        self._test_object_type(odl_const.ODL_SUBNET)

    def test_subnet_pending_network(self):
        self._test_object_type_pending_network(odl_const.ODL_SUBNET)

    def test_subnet_processing_network(self):
        self._test_object_type_processing_network(odl_const.ODL_SUBNET)

    def test_port(self):
        self._test_object_type(odl_const.ODL_PORT)

    def test_port_pending_network(self):
        self._test_object_type_pending_network(odl_const.ODL_PORT)

    def test_port_processing_network(self):
        self._test_object_type_processing_network(odl_const.ODL_PORT)



# Copyright (c) 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import threading

from requests import exceptions

from oslo_config import cfg
from oslo_log import log as logging

from neutron.db import api as neutron_db_api

from networking_odl.common import client
from networking_odl.common import constants as odl_const
from networking_odl.common import filters
from networking_odl.db import db
from networking_odl.openstack.common._i18n import _LE
from networking_odl.openstack.common._i18n import _LI

LOG = logging.getLogger(__name__)


def call_thread_on_end(func):
    def new_func(obj, *args):
        func(obj, *args)
        obj.journal.start_odl_sync_thread()
    return new_func


class OpendaylightJournalThread(object):
    """Thread worker for the Opendaylight Journal Database."""
    def __init__(self):
        self.client = client.OpenDaylightRestClient.create_client()
        self._odl_sync_timeout = cfg.CONF.ml2_odl.sync_timeout
        self._row_retry_count = cfg.CONF.ml2_odl.retry_count
        self.event = threading.Event()
        self.start_odl_sync_thread()

    def start_odl_sync_thread(self):
        # Reset timer
        if hasattr(self, 'timer'):
            LOG.debug("Resetting thread timer")
            self.timer.cancel()
            self.timer = None
        self.timer = threading.Timer(self._odl_sync_timeout,
                                     self.start_odl_sync_thread)
        self.timer.start()

        # Don't start a second thread if there is one alive already
        # Only trigger the event
        if (hasattr(self, '_odl_sync_thread') and
           self._odl_sync_thread.isAlive()):
            LOG.debug("Thread alive, sending event")
            self.event.set()
            return

        # Start the sync thread if there isn't one
        LOG.debug("Starting a new sync thread")
        self._odl_sync_thread = threading.Thread(
            name='sync',
            target=self.sync_pending_row)

        self._odl_sync_thread.start()

    def _json_data(self, row):
        filter_cls = filters.FILTER_MAP[row.object_type]

        if row.operation == odl_const.ODL_CREATE:
            method = 'post'
            attr_filter = filter_cls.filter_create_attributes
            data = copy.deepcopy(row.data)
            urlpath = row.object_type + 's'
            attr_filter(data)
            to_send = {row.object_type: data}
        elif row.operation == odl_const.ODL_UPDATE:
            method = 'put'
            attr_filter = filter_cls.filter_update_attributes
            data = copy.deepcopy(row.data)
            urlpath = row.object_type + 's/' + row.object_uuid
            attr_filter(data)
            to_send = {row.object_type: data}
        elif row.operation == odl_const.ODL_DELETE:
            method = 'delete'
            data = None
            urlpath = row.object_type + 's/' + row.object_uuid
            to_send = None

        return method, urlpath, to_send

    def sync_pending_row(self, exit_after_run=False):
        # Block until all pending rows are processed
        session = neutron_db_api.get_session()
        while not self.event.is_set():
            self.event.wait()
            # Clear the event and go back to waiting after
            # the sync block exits
            self.event.clear()
            while True:
                LOG.debug("Thread walking database")
                row = db.get_oldest_pending_db_row_with_lock(session)
                if not row:
                    LOG.debug("No rows to sync")
                    break

                # Validate the operation
                validate_func = db.VALIDATION_MAP[row.object_type]
                valid = validate_func(session, row.object_uuid,
                                      row.operation, row.data)
                if not valid:
                    LOG.info(_LI("%(operation)s %(type)s %(uuid)s is not a "
                                 "valid operation yet, skipping for now"),
                             {'operation': row.operation,
                              'type': row.object_type,
                              'uuid': row.object_uuid})
                    continue

                LOG.info(_LI("Syncing %(operation)s %(type)s %(uuid)s"),
                         {'operation': row.operation, 'type': row.object_type,
                          'uuid': row.object_uuid})

                # Add code to sync this to ODL
                method, urlpath, to_send = self._json_data(row)

                try:
                    self.client.sendjson(method, urlpath, to_send)
                    db.update_processing_db_row_passed(session, row)
                except exceptions.ConnectionError as e:
                    # Don't raise the retry count, just log an error
                    LOG.error(_LE("Cannot connect to the Opendaylight "
                                  "Controller"))
                    # Set row back to pending
                    db.update_db_row_pending(session, row)
                    # Break our of the loop and retry with the next
                    # timer interval
                    break
                except Exception as e:
                    LOG.error(_LE("Error syncing %(type)s %(operation)s,"
                                  " id %(uuid)s Error: %(error)s"),
                              {'type': row.object_type,
                               'uuid': row.object_uuid,
                               'operation': row.operation,
                               'error': e.message})
                    db.update_pending_db_row_retry(session, row,
                                                   self._row_retry_count)
            LOG.debug("Clearing sync thread event")
            if exit_after_run:
                # Permanently waiting thread model breaks unit tests
                # Adding this arg to exit here only for unit tests
                break



# Copyright (c) 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

import sqlalchemy as sa

from neutron.db import model_base
from neutron.db.models_v2 import HasId


class OpendaylightJournal(model_base.BASEV2, HasId):
    __tablename__ = 'opendaylightjournal'

    object_type = sa.Column(sa.String(36), nullable=False)
    object_uuid = sa.Column(sa.String(36), nullable=False)
    operation = sa.Column(sa.String(36), nullable=False)
    data = sa.Column(sa.PickleType, nullable=True)
    state = sa.Column(sa.Enum('pending', 'failed', 'processing', 'completed'),
                      nullable=False, default='pending')
    retry_count = sa.Column(sa.Integer, default=0)
    created_at = sa.Column(sa.DateTime)
    last_retried = sa.Column(sa.TIMESTAMP, nullable=False,
                             default=datetime.datetime.utcnow())
